from libpysal.weights.spatial_lag import lag_spatial as slag
import numpy as np
import random
from sklearn import preprocessing

PERMUTATIONS = 0


def get_memory():
    with open('/proc/meminfo', 'r') as mem:
        free_memory = 0
        for i in mem:
            sline = i.split()
            if str(sline[0]) in ('MemFree:', 'Buffers:', 'Cached:'):
                free_memory += int(sline[1])
    return free_memory / 1024


class Moran(object):
    """Moran's I Global Autocorrelation Statistic

    Modification from https://github.com/pysal/esda

    Parameters
    ----------

        n               : int
                          number of observations

        w               : W
                          spatial weights instance

        transformation  : string
                          weights transformation,  default is row-standardized "r".
                          Other options include "B": binary,  "D":
                          doubly-standardized,  "U": untransformed
                          (general weights), "V": variance-stabilizing.

        permutations    : int
                          number of random permutations for calculation of
                          pseudo-p_values

    Attributes
    ----------
        w            : W
                       original w object

        permutations : int
                       number of permutations

        I            : array
                       value of Moran's I

        sim          : array
                       (if permutations>0)
                       vector of I values for permuted samples

        p_sim        : array
                       (if permutations>0)
                       p-value based on permutations (one-tailed)
                       null: spatial randomness
                       alternative: the observed I is extreme if
                       it is either extremely greater or extremely lower
                       than the values obtained based on permutations

    Notes
    -----
    Technical details and derivations can be found in :cite:`cliff81`.

    """

    def __init__(
        self, n, w, transformation="r", permutations=PERMUTATIONS
    ):
        #y = np.asarray(y).flatten()
        #self.y = y
        w.transform = transformation
        self.w = w
        self.permutations = permutations
        self.n = n
        self.__moments()


    def calc_i(self, X, seed=1, max_RAM=16):
        """
        Function to calculate Moran's I of all features

        Arguments
        ---------
            X: n x p array
               log-transformed feature matrix

            seed: random seed to make the results reproducible

            max_RAM: maximum limitation of memory (Gb)

        Returns
        ---------
            the fitted estimator.

        """
        random.seed(seed)
        np.random.seed(seed)
        self.Z = preprocessing.StandardScaler().fit_transform(X)
        #self.Z = np.asarray(self.Z).flatten()
        self.Z[self.Z>10] = 10.0
        self.Z[self.Z<-10] = -10.0
        self.I = self.__calc(self.Z, max_RAM)

        if self.permutations:
            permutations = self.permutations
            sim = [
                self.__calc(np.random.permutation(self.Z), max_RAM) for i in range(permutations)
            ]

            self.sim = sim = np.array(sim)

            above = self.sim >= np.repeat(self.I[np.newaxis,:],self.permutations,axis=0)  #self.association_
            larger = above.sum(axis=0)
            extreme = np.minimum(self.permutations - larger, larger)
            self.p_sim = (extreme + 1.0) / (self.permutations + 1.0)

            #above = sim >= self.I
            #larger = above.sum()
            #if (self.permutations - larger) < larger:
            #    larger = self.permutations - larger
            #self.p_sim = (larger + 1.0) / (permutations + 1.0)


    def __moments(self):

        n = self.n
        n2 = n * n
        s1 = self.w.s1
        s0 = self.w.s0
        s2 = self.w.s2
        s02 = s0 * s0


    def __calc(self, Z, max_RAM):
        '''
        memory taken: 0.000008 * (3*n*p + 3*p)
        '''
        free_memory = min(get_memory(), max_RAM*1024)
        if (free_memory*0.8) > (0.000008 * (3 * Z.shape[1] * Z.shape[0] + 3*Z.shape[1])):
            z2ss = np.square(Z).sum(axis=0)
            zl = slag(self.w, Z)
            inum = (Z * zl).sum(axis=0)
            return self.n / self.w.s0 * (inum / z2ss)
        else:
            out = np.zeros(Z.shape[1])
            nt = (0.000008 * (3 * Z.shape[1] * Z.shape[0] + 3*Z.shape[1])) / (free_memory*0.8) + 1
            nt = int(nt)
            nf = Z.shape[1] // nt

            for i in range(nt-1):
                z2ss = np.square(self.Z[:,(nf*i):(nf*(i+1))]).sum(axis=0)
                zl = slag(self.w, Z[:,(nf*i):(nf*(i+1))])
                inum = (Z[:,(nf*i):(nf*(i+1))] * zl).sum(axis=0)
                out[(nf*i):(nf*(i+1))] = self.n / self.w.s0 * (inum / z2ss)

            z2ss = np.square(self.Z[:,(nf*(nt-1)):]).sum(axis=0)
            zl = slag(self.w, Z[:,(nf*(nt-1)):])
            inum = (Z[:,(nf*(nt-1)):] * zl).sum(axis=0)
            out[(nf*(nt-1)):] = self.n / self.w.s0 * (inum / z2ss)

            return out

        #z2ss = np.square(self.Z).sum(axis=0)
        #zl = slag(self.w, Z)
        #inum = (Z * zl).sum(axis=0)
        





import numpy
from scipy import sparse
from sklearn.base import BaseEstimator
from sklearn import preprocessing
from sklearn import utils
import random
from sys import getsizeof
import scipy.stats


def get_memory():
    with open('/proc/meminfo', 'r') as mem:
        free_memory = 0
        for i in mem:
            sline = i.split()
            if str(sline[0]) in ('MemFree:', 'Buffers:', 'Cached:'):
                free_memory += int(sline[1])
    return free_memory / 1024


class Spatial_Pearson(BaseEstimator):
    """Global Spatial Pearson Statistic"""

    def __init__(self, connectivity=None, permutations=999):
        """
        Initialize a spatial pearson estimator

        Arguments
        ---------
            connectivity:   scipy.sparse matrix object
                            the connectivity structure describing the relationships
                            between observed units. Will be row-standardized. 

            permutations:   int
                            the number of permutations to conduct for inference.
                            if < 1, no permutational inference will be conducted. 

        Attributes
        ----------
            association_: numpy.ndarray (p,)
                          array containg the estimated Lee spatial pearson correlation
                          coefficients for all gene-peak pairs

            reference_distribution_: numpy.ndarray (n_permutations, p)
                          distribution of correlation matrices for randomly-shuffled
                          maps. 

            significance_: numpy.ndarray (p,)
                           permutation-based z-scores (p-values) for the probability that
                           observed correlation was more extreme than the simulated 
                           correlations.
        """
        self.connectivity = connectivity
        self.permutations = permutations
        if self.connectivity is None:
            self.connectivity = sparse.eye(Z.shape[0])

        self.standard_connectivity = sparse.csc_matrix(
            self.connectivity / (self.connectivity.sum(axis=1)[:, numpy.newaxis])
        )

        self.ctc = self.connectivity.T @ self.connectivity
        ones = numpy.ones(self.ctc.shape[0])
        self.denom = (ones.T @ self.ctc @ ones)


    def fit(self, X, Y, percent=0.1, max_RAM=16, seed=1):
        """
        bivariate spatial pearson's R based on Eq. 18 of :cite:`Lee2001`.

        Arguments
        ---------
            X       :   numpy.ndarray [n x p]
                        array containing continuous data
            Y       :   numpy.ndarray [n x p]
                        array containing continuous data

            percent: percentage of cells to shuffle during permutation.
                     For most of the time, default 0.1 is alread a good choice.
            
            seed: random seed to make the results reproducible

            max_RAM: maximum limitation of memory (Gb)


        Returns
        -------
            the fitted estimator.

        Notes
        -----
            Technical details and derivations can be found in :cite:`Lee2001`.

        """
        random.seed(seed)
        numpy.random.seed(seed)
        X = utils.check_array(X)
        Y = utils.check_array(Y)
        X = preprocessing.StandardScaler().fit_transform(X)
        Y = preprocessing.StandardScaler().fit_transform(Y)
        X[X>10] = 10.0
        X[X<-10] = -10.0
        Y[Y>10] = 10.0
        Y[Y<-10] = -10.0
        

        self.association_ = self._statistic(X, Y, self.ctc, self.denom, max_RAM)


        if self.permutations is None:
            return self
        elif self.permutations < 1:
            return self

        if self.permutations:
            self.reference_distribution_ = numpy.zeros((self.permutations, X.shape[1]))

            for i in range(self.permutations):
                itp = numpy.random.choice(numpy.arange(X.shape[0]), int(X.shape[0]*percent), replace=False)
                rnd_index = numpy.arange(X.shape[0])
                rnd_index[itp] = numpy.random.permutation(itp)
                self.reference_distribution_[i] = self._statistic(Y[rnd_index], X[rnd_index], self.ctc, self.denom, max_RAM)

            self.z_sim = (self.association_ - numpy.mean(self.reference_distribution_, axis=0)) / numpy.std(self.reference_distribution_, axis=0)

            #above = self.reference_distribution_ >= numpy.repeat(self.association_[numpy.newaxis,:],self.permutations,axis=0)  #self.association_
            #larger = above.sum(axis=0)
            #extreme = numpy.minimum(self.permutations - larger, larger)
            #self.significance_ = (extreme + 1.0) / (self.permutations + 1.0)
            self.significance_ = scipy.stats.norm.sf(numpy.abs(self.z_sim))
        return self

    @staticmethod
    def _statistic(X, Y, ctc, denom, max_RAM):

        '''
        Memory taken (Mb): 0.000008 * (2 * p * n + p)
        '''
        free_memory = min(get_memory(), max_RAM*1024)
        if (free_memory*0.8) > (0.000008 * (2 * X.shape[1] * X.shape[0] + X.shape[1])):
            out = ((Y.T @ ctc) * X.T).sum(-1) / denom
        else:
            nt = (0.000008 * (2 * X.shape[1] * X.shape[0] + X.shape[1])) / (free_memory*0.8) + 1
            nt = int(nt)
            nf = X.shape[1] // nt
            out = numpy.zeros(X.shape[1])
            for i in range(nt-1):
                out[(nf*i):(nf*(i+1))] = ((Y[:,(nf*i):(nf*(i+1))].T @ ctc) * X[:,(nf*i):(nf*(i+1))].T).sum(-1) / denom

            out[(nf*(nt-1)):] = ((Y[:,(nf*(nt-1)):].T @ ctc) * X[:,(nf*(nt-1)):].T).sum(-1) / denom

        return out


class Spatial_Pearson_Local(BaseEstimator):
    """Local Spatial Pearson Statistic"""

    def __init__(self, connectivity=None, permutations=999):
        """
        Initialize a spatial local pearson estimator

        Arguments
        ---------
            connectivity:   scipy.sparse matrix object
                            the connectivity structure describing the relationships
                            between observed units. Will be row-standardized. 

            permutations:   int
                            the number of permutations to conduct for inference.
                            if < 1, no permutational inference will be conducted.


        Attributes
        ----------
            associations_: numpy.ndarray (n_samples,)
                          array containg the estimated Lee spatial pearson correlation
                          coefficients, where element [0,1] is the spatial correlation
                          coefficient, and elements [0,0] and [1,1] are the "spatial
                          smoothing factor"

            significance_: numpy.ndarray (n,p)
                           permutation-based z-scores (p-values) for the probability that
                           observed correlation was more extreme than the simulated 
                           correlations.


        Notes
        -----
        Technical details and derivations can be found in :cite:`Lee2001`.
        """
        self.connectivity = connectivity
        self.permutations = permutations

        self.standard_connectivity = sparse.csc_matrix(
                                        self.connectivity / (self.connectivity.sum(axis=1)[:, numpy.newaxis])
                                        )

    def fit(self, X, Y, max_RAM=16, seed=1):
        """
        bivariate local pearson's R based on Eq. 22 in Lee (2001)

        Arguments
        ---------
            X       :   numpy.ndarray [n x p]
                        array containing continuous data
            Y       :   numpy.ndarray [n x p]
                        array containing continuous data
            
            seed: random seed to make the results reproducible

            max_RAM: maximum limitation of memory (Gb)

        Returns
        -------
            the fitted estimator.
        """
        random.seed(seed)
        numpy.random.seed(seed)
        X = utils.check_array(X)
        X = preprocessing.StandardScaler().fit_transform(X)

        Y = utils.check_array(Y)
        Y = preprocessing.StandardScaler().fit_transform(Y)

        X[X>10] = 10.0
        X[X<-10] = -10.0
        Y[Y>10] = 10.0
        Y[Y<-10] = -10.0

        #Z = numpy.column_stack((x, y))

        n, _ = X.shape

        self.associations_ = self._statistic(X, Y, self.standard_connectivity, max_RAM)

        if self.permutations:

            self.significance_ = numpy.empty(X.shape)

            #perm_X = numpy.empty((self.permutations, X.shape[0], X.shape[1]))
            #perm_Y = numpy.empty((self.permutations, Y.shape[1], Y.shape[0]))

            for i in range(X.shape[1]):
                perm_X = numpy.empty((self.permutations, X.shape[0]))
                perm_Y = numpy.empty((self.permutations, Y.shape[0]))
                for j in range(self.permutations):
                    rnd_index = numpy.random.permutation(numpy.arange(X.shape[0]))
                    perm_X[j] = X[rnd_index,i]
                    perm_Y[j] = Y[rnd_index,i]

                reference_distribution = numpy.transpose(
                    (numpy.transpose(perm_Y[:,:,numpy.newaxis], (0,2,1)) @ self.standard_connectivity.T), (0, 2, 1)
                    ) * (self.standard_connectivity @ perm_X[:,:,numpy.newaxis])

                above = reference_distribution.squeeze(-1) >= numpy.repeat(self.associations_[:,i][numpy.newaxis,:],self.permutations,axis=0)
                larger = above.sum(axis=0)
                extreme = numpy.minimum(larger, self.permutations - larger)
                self.significance_[:,i] = (extreme + 1.0) / (self.permutations + 1.0)


        else:
            self.reference_distribution_ = None
        return self

    @staticmethod
    def _statistic(X, Y, W, max_RAM):

        '''
        Memory taken (Mb): 0.000008 * (3 * p * n)
        '''
        free_memory = min(get_memory(), max_RAM*1024)
        if (free_memory*0.8) > (0.000008 * (3 * X.shape[1] * X.shape[0])):
            return (Y.T @ W.T).T * (W @ X)
        else:
            nt = (0.000008 * (3 * X.shape[1] * X.shape[0])) / (free_memory*0.8) + 1
            nt = int(nt)
            nf = X.shape[1] // nt
            out = numpy.zeros(X.shape)
            for i in range(nt-1):
                out[:,(nf*i):(nf*(i+1))] = (Y[:,(nf*i):(nf*(i+1))].T @ W.T).T * (W @ X[:,(nf*i):(nf*(i+1))])

            out[:,(nf*(nt-1)):] = (Y[:,(nf*(nt-1)):].T @ W.T).T * (W @ X[:,(nf*(nt-1)):])

            return out





import numpy as np
import pandas as pd
import anndata as ad
import time

import treasmo.core as core
from minisom import MiniSom


from scipy import stats
import matplotlib.pyplot as plt
from statsmodels.stats.multitest import fdrcorrection
from sklearn.gaussian_process import GaussianProcessRegressor

import random
import scanpy as sc

import os
import sys
from sklearn import preprocessing

__author__ = "Chaozhong Liu"
__email__ = "czliubioinfo@gmail.com"


#=========================================================================================
# Cluster / Group marker identification
# Trajectory-based regulatory dynamics
# TF regulatory dynamics
# Statistical test for dynamics
#=========================================================================================




#=========================================================================================
# Group marker identification
#=========================================================================================

def FindAllMarkers(mudata, ident, mods=['rna','atac'], corrct_method='bonferroni', seed=1):
    """
    Function to discover regulatory markers in all groups

    Arguments
    ---------
        mudata: single-cell multi-omics data saved as MuData object
        
        ident: column name in mudata.obs containing group labels

        mods: scRNA-seq and scATAC-seq modality name in MuData object

        corrct_method: multi-test correction method, one of ['bonferroni', 'fdr']

        seed: random seed to make the results reproducible

    Returns
    ---------

        DataFrame - Differentially regulated pairs statistical test results


    """

    if 'Local_L' in mudata.uns.keys():
        local_L_df = pd.DataFrame(mudata.uns['Local_L'])
        local_L_df.columns = mudata.uns['Local_L_names']
    else:
        raise Exception("No previously calculated local correlation matrix found. \nPlease calculate the matrix first.")
        #mudata = core.Local_L(mudata, genes, peaks,
        #                      mods=mods,
        #                      rm_dropout=False,
        #                      seed=1, max_RAM=16)

        #local_L_df = pd.DataFrame(mudata.uns['Local_L'])
        #local_L_df.columns = mudata.uns['Local_L_names']
        #print("========= Finished ========")


    # cluster comparison
    print("Performing statistical test for correlation differences among identities...")
    start = time.time()
    # get clusters have at least 2 mini-bulk
    groups = mudata.obs[ident].value_counts() > 1
    groups = groups.loc[groups].index.to_list()

    local_L_df['clus'] = mudata.obs[ident].to_numpy()

    stat_df_all = pd.DataFrame(np.empty((0,13)))
    stat_df_all.columns = ['group','name','Mean.1','Mean.2','Std.1','Std.2',
                           'obsn.1','obsn.2','score','p.value','p.adj','Frac.gene.1','Frac.peak.1']

    for group in groups:
        stat_df = marker_test(local_L_df, group_1=group, group_2=None, corrct_method=corrct_method)
        try:
            stat_df['Frac.gene.1'], stat_df['Frac.peak.1'] = add_feature_sparsity(stat_df, mudata, mods=mods, group=group)
        except:
            print('\tCluster %s sparsity information not found. Skip.'%group)
            stat_df['Frac.gene.1'] = np.NaN
            stat_df['Frac.peak.1'] = np.NaN

        stat_df_all = pd.concat([stat_df_all,stat_df])
    
    print("Completed! %.2fs past."%(time.time()-start))

    return stat_df_all



def FindMarkers(mudata, ident, group_1, group_2, mods=['rna','atac'], corrct_method='bonferroni',seed=1, log=True):
    """
    Function to compare regulatory pairs between two group

    Arguments
    ---------
        mudata: single-cell multi-omics data saved as MuData object
        
        ident: column name in mudata.obs containing group labels

        group_1: first group name in ident to compare with the second

        group_2: second group name in ident to compare with the first

        mods: scRNA-seq and scATAC-seq modality name in MuData object

        corrct_method: multi-test correction method, one of ['bonferroni', 'fdr']

        seed: random seed to make the results reproducible

    Returns
    ---------

        DataFrame - Differentially regulated pairs statistical test results


    """

    if 'Local_L' in mudata.uns.keys():
        local_L_df = pd.DataFrame(mudata.uns['Local_L'])
        local_L_df.columns = mudata.uns['Local_L_names']
    else:
        raise Exception("No previously calculated local correlation matrix found. \nPlease calculate the matrix first.")
        #mudata = core.Local_L(mudata, genes, peaks,
        #                      mods=mods,
        #                      rm_dropout=False,
        #                      seed=1, max_RAM=16)

        #local_L_df = pd.DataFrame(mudata.uns['Local_L'])
        #local_L_df.columns = mudata.uns['Local_L_names']
        #print("========= Finished ========")


    # cluster comparison
    if log:
        print("Perform statistical test for correlation differences between selected two group...")
    start = time.time()

    local_L_df['clus'] = mudata.obs[ident].to_numpy()
    stat_df = marker_test(local_L_df, group_1=group_1, group_2=group_2, corrct_method=corrct_method)

    try:
        stat_df['Frac.gene.1'], stat_df['Frac.peak.1'] = add_feature_sparsity(stat_df, mudata, mods=mods, group=group_1)
        stat_df['Frac.gene.2'], stat_df['Frac.peak.2'] = add_feature_sparsity(stat_df, mudata, mods=mods, group=group_2)
    except:
        if log:
            print('\tSparsity information not found. Skip.')
        stat_df['Frac.gene.1'] = np.NaN
        stat_df['Frac.peak.1'] = np.NaN
        stat_df['Frac.gene.2'] = np.NaN
        stat_df['Frac.peak.2'] = np.NaN

    if log:
        print("Completed! %.2fs past."%(time.time()-start))

    return stat_df




def marker_test(local_L_df, group_1, group_2=None, corrct_method='bonferroni'):

    #stat_df.columns = ['group','name','Mean.1','Mean.2','Std.1',
    #                       'Std.2','obsn.1','obsn.2','score','p.value','p.adj']

    stat_df = group_stat(local_L_df, group_1=group_1, group_2=group_2)
    stat_df['score'] = np.NaN
    stat_df['p.value'] = np.NaN
    for i in range(stat_df.shape[0]):
        ttest = stats.ttest_ind_from_stats(
                mean1=stat_df['Mean.1'][i],
                std1=stat_df['Std.1'][i],
                nobs1=stat_df['obsn.1'][i],
                mean2=stat_df['Mean.2'][i],
                std2=stat_df['Std.2'][i],
                nobs2=stat_df['obsn.2'][i],
                equal_var=True,  # Welch's
                )
        stat_df.loc[stat_df.index[i],'score'] = ttest[0]
        stat_df.loc[stat_df.index[i],'p.value'] = ttest[1]
    if corrct_method == 'fdr':
        _, stat_df['p.adj'] = fdrcorrection(
                    stat_df['p.value'], alpha=0.05, method='indep'
                )
    elif corrct_method == 'bonferroni':
        stat_df['p.adj'] = np.minimum(stat_df['p.value'] * stat_df.shape[0], 1.0)
    else:
        #print("Please select correction methods from ['fdr', 'bonferroni']!")
        stat_df['p.adj'] = np.NaN

    #stat_df['Frac.gene'], stat_df['Frac.peak'] = _add_clus_sparsity(stat_df, anndat_multiome, group)

    #stat_df_all = pd.concat([stat_df_all,stat_df])

    return stat_df


def group_stat(local_L, group_1, group_2=None):
    groups_label = local_L['clus'].to_numpy()
    mask_1 = groups_label == group_1
    obsn1 = np.sum(mask_1)
    if group_2 is None:
        group_2 = 'others'
        mask_2 = groups_label != group_1
        obsn2 = len(mask_2) - obsn1
    else:
        mask_2 = groups_label == group_2
        obsn2 = np.sum(mask_2)

    local_L_df = local_L.iloc[:,:-1].copy()
    local_L_df['clus'] = 'others'
    local_L_df.loc[mask_1, 'clus'] = group_1
    local_L_df.loc[mask_2, 'clus'] = group_2

    mean_df = local_L_df.groupby('clus').mean() #.reset_index()
    mean_df = mean_df.loc[[group_1,group_2],:]
    std_df = local_L_df.groupby('clus').std()
    std_df = std_df.loc[[group_1,group_2],:]

    mean_df = mean_df.T
    mean_df.columns = ['Mean.1','Mean.2']

    std_df = std_df.T
    std_df.columns = ['Std.1','Std.2']

    stat_df = pd.concat([mean_df, std_df], axis=1)
    stat_df.insert(0,'name',stat_df.index)
    stat_df = stat_df.reset_index(drop=True)

    if group_2 == 'others':
        stat_df.insert(0,'group',group_1)
    else:
        stat_df.insert(0,'group.1',group_1)
        stat_df.insert(1,'group.2',group_2)

    stat_df['obsn.1'] = obsn1
    stat_df['obsn.2'] = obsn2
    
    return stat_df


def add_feature_sparsity(stat_df, mudata, group, mods=['rna','atac']):
    GP_G = [gp.split('~')[0] for gp in stat_df['name']]
    GP_P = [gp.split('~')[1] for gp in stat_df['name']]
    
    return mudata.mod[mods[0]].var.loc[GP_G,:]['Frac.%s'%group].to_numpy(),\
           mudata.mod[mods[1]].var.loc[GP_P,:]['Frac.%s'%group].to_numpy()




def MarkerFilter(statDf, min_pct_rna=0.1, min_pct_atac=0.05, mean_diff=1.0, p_cutoff=1e-12, plot=False):
    """
    Function to filter markers from statistical test results by sparsity, correlation difference, and p-value

    Arguments
    ---------
        statDf: Differentially regulated pairs statistical test results
        
        min_pct_rna: percentage of cells that express the gene as sparsity cutoff

        min_pct_atac: percentage of cells that have the peak as sparsity cutoff

        mean_diff: mean correlation difference between the group and background (all other groups)

        p_cutoff: adjusted p-value cutoff

        plot: if True, return volcano plot

    Returns
    ---------
        Filtered marker list with the same columns as stat_df

        if plot==True, return volcano plot


    """
    stat_df = statDf.copy()

    pd.options.mode.chained_assignment = None
    stat_df.loc[:,'score.abs'] = np.abs(stat_df['score'])
    if 'group' in stat_df.columns:
        stat_df = stat_df.sort_values(by=['group','score.abs'], ascending=False)
    else:
        stat_df = stat_df.sort_values(by=['group.1','score.abs'], ascending=False)

    mask = (stat_df['Frac.gene.1']>min_pct_rna) & (stat_df['Frac.peak.1']>min_pct_atac)
    stat_df = stat_df.loc[mask,:]

    filt = ((stat_df['p.adj']<p_cutoff) &
        (np.abs(stat_df['Mean.1']-stat_df['Mean.2']) > mean_diff))

    if plot:
        plt.scatter(stat_df['Mean.1'][~filt]-stat_df['Mean.2'][~filt],
                    -np.log10(stat_df['p.adj'][~filt]), s=2, marker='o', c='grey')
        plt.scatter(stat_df['Mean.1'][filt]-stat_df['Mean.2'][filt],
                    -np.log10(stat_df['p.adj'][filt]), s=2, marker='o', c='red')
        plt.grid(which='both', linestyle='-', linewidth='0')
        plt.xlabel('mean_1 - mean_2')
        plt.ylabel('-log10(p.adj)')

    return stat_df.loc[filt,:]




#=========================================================================================
# Trajectory Path Analysis
#=========================================================================================


def FindPathMarkers(mudata, ident, path, mods=['rna','atac'], corrct_method='bonferroni', seed=1):
    """
    One-to-one gene-peak pair correlation comparison among groups in the trajectory path

    Arguments
    ---------
        mudata: single-cell multi-omics data saved as MuData object
        
        ident: column name in mudata.obs containing group labels

        path: list of clusters ordered by their sequence on the trajectory. A path here should have no branch.

        mods: scRNA-seq and scATAC-seq modality name in MuData object

        corrct_method: multi-test correction method, one of ['bonferroni', 'fdr']

        seed: random seed to make the results reproducible

    Returns
    ---------

        DataFrame - Differentially regulated pairs statistical test results


    """

    # One-to-one comparison
    start = time.time()

    dfList = []

    for i in range(len(path)):

        for j in range(i+1, len(path)):

            stat_df = FindMarkers(mudata, ident, 
                                  group_1=path[j], group_2=path[i], 
                                  mods=mods, 
                                  corrct_method=corrct_method, 
                                  seed=seed, log=False)
            dfList.append(stat_df)

    statDf = pd.concat(dfList)

    print("Completed! %.3fs past."%(time.time()-start))

    return statDf




def _fit_data(timebinDf, xfit):
    xdata = timebinDf['time'].to_numpy()
    ydata = timebinDf['value'].to_numpy()

    filter_na = ~np.isnan(ydata)
    xdata = xdata[filter_na]
    ydata = ydata[filter_na]

    # Compute the Gaussian process fit
    gp = GaussianProcessRegressor(random_state=1)
    gp.fit(xdata[:, np.newaxis], ydata)
    #xfit = np.linspace(data['time'].min(), data['time'].max(), bins)
    yfit, _ = gp.predict(xfit[:, np.newaxis], return_std=True)

    return yfit



def TimeBinData(mudata, ident, path, pseudotime, features,
                bins=100, rm_outlier=False, fitted=None):

    cells_bool = mudata.obs[ident].isin(path).to_numpy()
    #features = mudata.uns['Local_L_names']
    CorDf = pd.DataFrame(mudata.uns['Local_L'], columns=mudata.uns['Local_L_names'])
    CorDf = CorDf.loc[cells_bool, features]

    CorDf['time'] = mudata.obs.loc[cells_bool, pseudotime].to_numpy()
    CorDf = CorDf.sort_values(by='time')

    # Construct time bins
    time_range = (CorDf['time'].min(), CorDf['time'].max())
    N_interval = bins
    interv_range = (time_range[1] - time_range[0]) / N_interval

    group_intev = pd.cut(CorDf['time'], np.arange(time_range[0]-1e-3, 
                                                  time_range[1]+interv_range,
                                                  interv_range))

    data = CorDf.groupby(group_intev).mean().dropna()
    print("Empty bins removed. %.i bins left"%(data.shape[0]))
    data.index = np.arange(data.shape[0])
    data.index.name = None

    if rm_outlier:
        data = data[np.abs(stats.zscore(data)) <= 2]

    data = data.sort_values(by='time')

    if fitted is not None:
        fitDf = pd.DataFrame(np.zeros((fitted, data.shape[1])), columns=data.columns)
        timefit = np.linspace(data['time'].min(), data['time'].max(), fitted)
        fitDf['time'] = timefit

        for feature in features:
            tmpDf = data[['time', feature]].copy()
            tmpDf.columns = ['time', 'value']
            fitDf[feature] = _fit_data(tmpDf, timefit)

        return data, fitDf

    else:
        return data




def TimeBinProportion(mudata, ident, path, pseudotime, bins=100):

    # Construct DataFrame
    cells_bool = mudata.obs[ident].isin(path).to_numpy()
    CorDf = pd.DataFrame(pd.get_dummies(mudata.obs.loc[cells_bool, ident]).astype('float32').loc[:,path].to_numpy(),
                         columns=path)
    CorDf['time'] = mudata.obs.loc[cells_bool, pseudotime].to_numpy()

    CorDf = CorDf.sort_values(by='time')

    # Construct time bins
    time_range = (CorDf['time'].min(), CorDf['time'].max())
    N_interval = bins
    interv_range = (time_range[1] - time_range[0]) / N_interval

    group_intev = pd.cut(CorDf['time'], np.arange(time_range[0]-1e-3, 
                                                  time_range[1]+interv_range,
                                                  interv_range))

    data = CorDf.groupby(group_intev).mean().dropna()
    print("Empty bins removed. %.i bins left"%(data.shape[0]))
    data.index = np.arange(data.shape[0])
    data.index.name = None
    data.columns.name = ident

    return data





def FindPathDynamics(mudata, ident, path, pseudotime, rm_outlier=True,
                     var_cutoff=0.1, range_cutoff=1.0, bins=100, plot=False):
    """
    Detect highly variable gene-peak pairs along the trajectory
    """
    # Construct DataFrame
    features = mudata.uns['Local_L_names']
    data = TimeBinData(mudata, ident, path, pseudotime, features,
                       bins=bins, rm_outlier=True)

    sumDf = pd.DataFrame(data[features].var(), columns=['variance'])
    sumDf['max'] = data[features].max()
    sumDf['min'] = data[features].min()
    sumDf['range'] = sumDf['max'] - sumDf['min']
    sumDf = sumDf.sort_values(by='variance',ascending=False)

    filt = ((sumDf['variance']>=var_cutoff) &(sumDf['range']>=range_cutoff))

    if plot:
        plt.scatter(sumDf['range'][~filt],
                    sumDf['variance'][~filt], s=2, marker='o', c='grey')
        plt.scatter(sumDf['range'][filt],
                    sumDf['variance'][filt], s=2, marker='o', c='red')
        plt.grid(which='both', linestyle='-', linewidth='0')
        plt.xlabel('min - max')
        plt.ylabel('variance')

    return sumDf[filt]





def PathDynamics(mudata, gene, peaks, ident, path, pseudotime, bins=100):
    """
    Quantify regulatory dynamics along the trajectory.

    Arguments
    ---------
        mudata: single-cell multi-omics data saved as MuData object
        
        gene: gene name

        peak: a list of peaks correlated with the gene (gene-peak pair should exist in mudata.uns['Local_L_names'])

        ident: ident label in mudata.obs columns to distinguish clusters listed in path

        path: list of clusters ordered by their sequence on the trajectory. A path here should have no branch.

        pseudotime: pseudotime label for the trajectory saved in mudata.obs

        bins: number of bins to divide the trajectory into


    Returns
    ---------

        DataFrame saved in mudata.uns['path_dynamics'] describing correlation and cluster proportion changes along the trajectory


    """

    features = [f"{gene}~{peak}" for peak in peaks]

    # Construct DataFrame
    cells_bool = mudata.obs[ident].isin(path).to_numpy()
    CorDf = pd.DataFrame(mudata.uns['Local_L'], columns=mudata.uns['Local_L_names'])
    CorDf = CorDf.loc[cells_bool, features]

    CorDf['time'] = mudata.obs.loc[cells_bool, pseudotime].to_numpy()
    CorDf[path] = pd.get_dummies(mudata.obs.loc[cells_bool, ident]).astype('float32').loc[:,path].to_numpy()

    CorDf = CorDf.sort_values(by='time')

    # Construct time bins
    time_range = (CorDf['time'].min(), CorDf['time'].max())
    N_interval = bins
    interv_range = (time_range[1] - time_range[0]) / N_interval

    group_intev = pd.cut(CorDf['time'], np.arange(time_range[0]-1e-3, 
                                                  time_range[1]+interv_range,
                                                  interv_range))
    '''
    agg_dict = {feature:['mean', 'std']}
    agg_dict['time'] = 'mean'
    for ct in path:
        agg_dict[ct] = 'mean'

    data = CorDf.groupby(group_intev).agg(agg_dict).dropna() #.mean().dropna()
    '''
    data = CorDf.groupby(group_intev).mean().dropna()

    print("Empty bins removed. %.i bins left"%(data.shape[0]))
    data.index = np.arange(data.shape[0])
    data.index.name = None

    path_name = '_'.join(path)

    if 'pathDym' not in mudata.uns.keys():
        mudata.uns['pathDym'] = {}

    if path_name not in mudata.uns['pathDym'].keys():
        mudata.uns['pathDym'][path_name] = {gene:data}
    else:
        mudata.uns['pathDym'][path_name][gene] = data
    
    return mudata



def DynamicModule(mudata, ident, path, pseudotime, features=None, bins=100, fitted=100,
                  num_iteration=5000, som_shape=(2,2), sigma=0.5, learning_rate=.1, random_seed=1):
    
    """
    Function to discover feature modules by Self-Organizing Map


    Arguments
    ---------
        mudata: single-cell multi-omics data saved as MuData object

        ident: ident label in mudata.obs columns to distinguish clusters listed in path

        path: list of clusters ordered by their sequence on the trajectory. A path here should have no branch.

        pseudotime: pseudotime label for the trajectory saved in mudata.obs

        bins: number of bins to divide the trajectory into

        re_outlier: whether to remove bins exceeding 2 std
    
        num_iteration: maximum number of iteration to optimize the SOM
        
        som_shape: (M, N) shape of the map, defines number and similarity structure of modules

        sigma: the radius of the different neighbors in the SOM

        learning_rate: optimization speed, how much weights are adjusted during each iteration

        random_seed: random seed to make the results reproducible


    Returns
    ---------
        time bin data labeled with module


    """
    if features is None:
        features = mudata.uns['Local_L_names']

    data, fitDf = TimeBinData(mudata, ident, path, pseudotime, features,
                              bins=bins, rm_outlier=False, fitted=fitted)


    som_mtx = fitDf.to_numpy()[:,:-1].T
    som_mtx = stats.zscore(som_mtx, axis=1)
    som_mtx[som_mtx > 3] = 3
    som_mtx[som_mtx < -3] = -3

    print("Start training...")
    som = MiniSom(som_shape[0], som_shape[1], som_mtx.shape[1], sigma=sigma, learning_rate=learning_rate,
                  neighborhood_function='gaussian', random_seed=random_seed)

    som.train_batch(som_mtx, num_iteration=num_iteration, verbose=True)

    # record module index
    winner_coordinates = np.array([som.winner(x) for x in som_mtx]).T
    cluster_index = np.ravel_multi_index(winner_coordinates, som_shape)
    somDict = {clsi:data.loc[:,np.concatenate([cluster_index==clsi, [True]])].copy() for clsi in np.unique(cluster_index)}

    return somDict







import numpy as np
import pandas as pd
import scanpy as sc
import anndata as ad

import scipy as sp
from scipy import stats
import matplotlib.pyplot as plt
import random
import time

from sklearn.neighbors import KNeighborsRegressor


__author__ = "Chaozhong Liu"
__email__ = "czliubioinfo@gmail.com"


# ============= Toolkit functions =============
# For Future work
#   Pseudo-bulk Generation (and the functions mapping results back to single-cell data)
#   nearest neighbor map construction
#   imputation option


# Implemented function
#   Feature sparsity summary
#   Gene-peak annotation
#   TF-peak annotation
# =============================================



# Calculate feature sparsity =================================================

def _fill_miss_index(df, n):
    miss_ind = np.setdiff1d(np.arange(n), df['ind'].to_numpy())
    miss_pd = pd.DataFrame({'ind':miss_ind,'size':0})
    df = pd.concat([df,miss_pd], axis=0).sort_values('ind')
    return df


def feature_sparsity(mudata, group_by={}):
    """
    Function to add sparisity info in AnnData.var

    Arguments
    --------------
        mudata: single-cell multi-omics data saved as MuData object

        group_by: dict; If provided, calculate per group feature sparsity for each modality
                  Example: {'rna':'cell_type', 'atac':'cell_type'}

    Return
    --------------
        anndat with
            var['Frac.all']
            var['Frac.GroupName']

    """

    for mod in mudata.mod.keys():

        spsmtx = sp.sparse.csr_matrix(mudata.mod[mod].X)
        non_zero = spsmtx.nonzero()
        non_zero_row = non_zero[0]
        non_zero_ind = non_zero[1]

        ## All cell sparsity
        non_zero_pd = pd.DataFrame(non_zero_ind)
        non_zero_pd = non_zero_pd.groupby(by=0, as_index=False).size()
        non_zero_pd = pd.DataFrame(non_zero_pd)
        non_zero_pd.columns = ['ind','size']
        non_zero_pd = _fill_miss_index(non_zero_pd, mudata.mod[mod].n_vars)
        non_zero_pd['sparsity'] = non_zero_pd['size'] / mudata.mod[mod].n_obs

        mudata.mod[mod].var['Frac.all'] = non_zero_pd['sparsity'].to_numpy()

        ## Per group sparsity
        if mod in group_by.keys():

            non_zero_pd = pd.DataFrame({'ind':non_zero_ind,
                                        'group':mudata.mod[mod].obs[group_by[mod]].to_numpy()[non_zero_row]})
            non_zero_pd = non_zero_pd.groupby(by=['ind','group'], as_index=False).size().sort_values('group')
            non_zero_pd['n_cells'] = mudata.mod[mod].obs[group_by[mod]].value_counts()[non_zero_pd['group']].to_numpy()

            for ident in mudata.mod[mod].obs[group_by[mod]].unique():
                df = non_zero_pd.loc[non_zero_pd['group']==ident,:].copy()
                n_cells = df['n_cells'].tolist()[0]
                df = df.loc[:,['ind','size']]
                df = _fill_miss_index(df, mudata.mod[mod].n_vars)
                df['sparsity'] = df['size'] / n_cells
                mudata.mod[mod].var['Frac.%s'%ident] = df['sparsity'].to_numpy()


    return mudata




# Annotate genes with nearby peaks =================================================

def get_gloc_from_atac_data(peaks, split_symbol):
    """
    Method to get the genomic locations (including the middle point)of peaks
        Author: Linhua Wang Linhua.Wang@bcm.edu
        https://github.com/LiuzLab/Neurips2021/blob/master/task1_utils1.py
    """
    glocs = peaks.tolist()
    glocs = [c for c in glocs if 'chr' in c]
    chrms, ranges, sts, ends, midpoints = [], [], [], [], []
    for gl in glocs:
        chrms.append(gl.split(split_symbol[0],1)[0])
        st, end = int(gl.split(split_symbol[0],1)[1].split(split_symbol[1],1)[0]), int(gl.split(split_symbol[0],1)[1].split(split_symbol[1],1)[1])
        sts.append(st)
        ends.append(end)
        #midpoints.append(int((st + end)/2))
        #ranges.append("_".join(gl.split(split_symbol[0])[1].split(split_symbol[1])))
    gloc_df = pd.DataFrame({'chrm': chrms, #'grange': ranges,
                        'start': sts, 'end': ends}, index=glocs)
                        #'midpoint': midpoints}, index=glocs)
    return gloc_df




def nearby_peaks(g_array, min_op=1):
    g_chr = str(g_array[2])
    g_array_n = g_array[0:2].copy().astype(float)
    op = np.minimum(p_array[:,1].squeeze(),g_array_n[1]) - np.maximum(p_array[:,0].squeeze(),g_array_n[0])
    filter_bool = (op>=min_op) & (chr_list==g_chr)
    op_index = np.arange(op.shape[0])[filter_bool]
    return np.array(','.join(list(plist[op_index])), dtype=object)
    

def peaks_within_distance(genes, peaks, upstream, downstream, ref_gtf_fn,
                          no_intersect=True, id_type='Symbol', split_symbol=['-','-']):
    """
    Function to annotate genes with nearby peaks

    Arguments
    --------------
        genes: gene list to be annotated

        peaks: peak candidate list

        upstream: include peaks N bp upstream of the TSS

        downstream: include peaks N bp downstream of the TES

        ref_gtf_fn: GTF format file containing gene location information
                    see example at https://github.com/ChaozhongLiu/scGREAT/tree/main/replication/data
                    - Homo_sapiens.GRCh38.104.GeneLoc.Tab.txt
                    - Mus_musculus.GRCm38.100.GeneLoc.Tab.txt

        no_intersect: if peak within the range lies in another gene body, remove the peak or not

        id_type: ID type of the gene list, ref_gtf_fn should contain this ID

        split_symbol: how peak location ID is merged
                      'chr1-12345-23456' - split_symbol=['-','-']
                      'chr1:12345-23456' - split_symbol=[':','-']

    Return
    --------------
        DataFrame containing gene annotation infos

    """

    gloc_df = get_gloc_from_atac_data(peaks, split_symbol=split_symbol)
    
    ref_gtf = pd.read_csv(ref_gtf_fn, sep='\t')
    if id_type=='Symbol':
        ref_gtf = ref_gtf.loc[ref_gtf['GeneSymbol'].isin(genes)]
    elif id_type=='Ensembl':
        ref_gtf = ref_gtf.loc[ref_gtf['gene_id'].isin(genes)]
        ref_gtf['GeneSymbol'] = ref_gtf['gene_id'].copy()
    #upstream, downstream = 100000, 100000
    ref_gtf['start_exp'] = 0
    ref_gtf['end_exp'] = 0

    ref_gtf.loc[ref_gtf['Strand']=='+','start_exp'] = ref_gtf.loc[ref_gtf['Strand']=='+','start'] - upstream
    ref_gtf.loc[ref_gtf['Strand']=='-','start_exp'] = ref_gtf.loc[ref_gtf['Strand']=='-','start'] - downstream

    ref_gtf.loc[ref_gtf['Strand']=='+','end_exp'] = ref_gtf.loc[ref_gtf['Strand']=='+','end'] + downstream
    ref_gtf.loc[ref_gtf['Strand']=='-','end_exp'] = ref_gtf.loc[ref_gtf['Strand']=='-','end'] + upstream
    
    global p_array, plist, chr_list
    p_array = gloc_df[['start','end']].to_numpy()
    plist = gloc_df.index.to_numpy()
    chr_list = gloc_df['chrm'].to_numpy().astype(object)

    genes_loc = ref_gtf[['start_exp','end_exp','chr']].to_numpy()
    selected_peaks = np.apply_along_axis(nearby_peaks, axis=1, arr=genes_loc)
    
    ref_gtf['nearby_peaks'] = selected_peaks
    ref_gtf = ref_gtf.loc[~(ref_gtf['nearby_peaks']=='')]
    ref_gtf['nearby_peaks'] = ref_gtf['nearby_peaks'].str.split(',')
    
    peaks_nearby_new = pd.DataFrame(np.ones((0,6)), columns=['GeneSymbol','Strand','chr','start','end','nearby_peaks']) #ref_gtf.columns[[1,4,5,6,7,10]])
    for i in range(ref_gtf.shape[0]):
        n_lines = len(ref_gtf.loc[ref_gtf.index[i],'nearby_peaks'])
        peaks_nearby_new_tmp = ref_gtf.iloc[np.repeat(i,n_lines),:]
        peaks_nearby_new_tmp = peaks_nearby_new_tmp.loc[:,['GeneSymbol','Strand','chr','start','end','nearby_peaks']]
        peaks_nearby_new_tmp['nearby_peaks'] = ref_gtf.loc[ref_gtf.index[i],'nearby_peaks']

        peaks_nearby_new = pd.concat([peaks_nearby_new, peaks_nearby_new_tmp], axis=0)
        peaks_nearby_new.index = np.arange(peaks_nearby_new.shape[0])
    
    #return peaks_nearby_new

    peaks_nearby_new['midp'] = peaks_nearby_new['nearby_peaks'].apply(lambda x: \
                               int((int(x.split(split_symbol[0],1)[1].split(split_symbol[1],1)[0]) + int(x.split(split_symbol[0],1)[1].split(split_symbol[1],1)[1]))/2) )

    plus_strand = peaks_nearby_new['Strand']=='+'

    peaks_nearby_new['tss_dist'] = np.NaN
    peaks_nearby_new.loc[plus_strand,'tss_dist'] = peaks_nearby_new.loc[plus_strand,'midp'] - peaks_nearby_new.loc[plus_strand,'start']
    peaks_nearby_new.loc[~plus_strand,'tss_dist'] = peaks_nearby_new.loc[~plus_strand,'end'] - peaks_nearby_new.loc[~plus_strand,'midp']

    peaks_nearby_new['tts_dist'] = np.NaN
    peaks_nearby_new.loc[plus_strand,'tts_dist'] = peaks_nearby_new.loc[plus_strand,'midp'] - peaks_nearby_new.loc[plus_strand,'end']
    peaks_nearby_new.loc[~plus_strand,'tts_dist'] = peaks_nearby_new.loc[~plus_strand,'start'] - peaks_nearby_new.loc[~plus_strand,'midp']

    promoters = ((peaks_nearby_new['tss_dist'] >= -2000) & (peaks_nearby_new['tss_dist'] <= 0)).astype(int)
    genebodys = ((peaks_nearby_new['tss_dist'] > 0) & (peaks_nearby_new['tts_dist'] <= 0)).astype(int)

    peaks_nearby_new['pRegion'] = promoters
    peaks_nearby_new['gBody'] = genebodys
    
    peaks_nearby_final = peaks_nearby_new.loc[(peaks_nearby_new['tss_dist']>=-upstream)&
                                              (peaks_nearby_new['tts_dist']<=downstream)].copy()
    
    df = peaks_nearby_final.copy()
    df.columns = ['Gene','Strand','chr','start','end','Peak','midp','tss_dist','tts_dist','pRegion','gBody']
    df = df.loc[~df.iloc[:,[0,5]].duplicated(),:].copy()
    
    if no_intersect:
        print("Remove nearby peaks if it lies on the gene body or promoter regions of other genes.")
        duplicated_peaks = df['Peak'].duplicated(keep=False)
        dup_df = df.loc[duplicated_peaks,:].copy()
        dup_df = dup_df.groupby(by='Peak')[['pRegion','gBody']].max()
        body_peaks = dup_df.index[(dup_df['pRegion']==1) | (dup_df['gBody']==1)]

        peaks_rm = (df['Peak'].isin(body_peaks)) & (df['pRegion']==0) & (df['gBody']==0)

        df = df.loc[~peaks_rm,:]
    
    df = df.iloc[:,[0,5,7,9,8,10]].copy()
    df.index = np.arange(df.shape[0])
    
    return df




# Annotate TF with motif regions =================================================


def TFBS_match(genes, peaks, ref_fn, min_overlap=1, split_symbol=['-','-']):

    """
    Function to annotate TF with binding sites

    Arguments
    --------------
        genes: gene list to be annotated

        peaks: peak candidate list

        ref_fn: BED format file containing gene binding site location information
                see example below from JASPAR TFBS genome track (https://jaspar.genereg.net/genome-tracks/)
                    chr1    280     298     AGL3    821     -
                    chr1    309     327     AGL3    823     +
                    chr1    309     327     AGL3    882     -
                    chr1    1577    1595    AGL3    823     +
                    chr1    1577    1595    AGL3    883     -
        
        min_overlap: number; the minimum number of overlaped base pairs between peak and TFBS

        split_symbol: how peak location ID is merged
                      'chr1-12345-23456' - split_symbol=['-','-']
                      'chr1:12345-23456' - split_symbol=[':','-']

    Return
    --------------
        DataFrame containing TF annotation infos

    """


    gloc_df = get_gloc_from_atac_data(peaks, split_symbol=split_symbol)
    
    ref_gtf = pd.read_csv(ref_fn, sep='\t', header=None)
    ref_gtf.columns = ['chr', 'start', 'end', 'Gene', 'score', 'Strand']
    ref_gtf = ref_gtf.loc[ref_gtf['Gene'].isin(genes)]

    upstream, downstream = 0, 0
    ref_gtf['start_exp'] = 0
    ref_gtf['end_exp'] = 0

    ref_gtf.loc[ref_gtf['Strand']=='+','start_exp'] = ref_gtf.loc[ref_gtf['Strand']=='+','start'] - upstream
    ref_gtf.loc[ref_gtf['Strand']=='-','start_exp'] = ref_gtf.loc[ref_gtf['Strand']=='-','start'] - downstream

    ref_gtf.loc[ref_gtf['Strand']=='+','end_exp'] = ref_gtf.loc[ref_gtf['Strand']=='+','end'] + downstream
    ref_gtf.loc[ref_gtf['Strand']=='-','end_exp'] = ref_gtf.loc[ref_gtf['Strand']=='-','end'] + upstream

    global p_array, plist, chr_list
    p_array = gloc_df[['start','end']].to_numpy()
    plist = gloc_df.index.to_numpy()
    chr_list = gloc_df['chrm'].to_numpy().astype(object)

    genes_loc = ref_gtf[['start_exp','end_exp','chr']].to_numpy()
    selected_peaks = np.apply_along_axis(nearby_peaks, axis=1, arr=genes_loc)
    
    ref_gtf['TFBS'] = selected_peaks
    ref_gtf = ref_gtf.loc[~(ref_gtf['TFBS']=='')]
    ref_gtf['TFBS'] = ref_gtf['TFBS'].str.split(',')

    peaks_nearby_new = pd.DataFrame(np.ones((0,6)), columns=['Gene','Strand','chr','start','end','TFBS']) #ref_gtf.columns[[1,4,5,6,7,10]])
    for i in range(ref_gtf.shape[0]):
        n_lines = len(ref_gtf.loc[ref_gtf.index[i],'TFBS'])
        peaks_nearby_new_tmp = ref_gtf.iloc[np.repeat(i,n_lines),:]
        peaks_nearby_new_tmp = peaks_nearby_new_tmp.loc[:,['Gene','Strand','chr','start','end','TFBS']]
        peaks_nearby_new_tmp['TFBS'] = ref_gtf.loc[ref_gtf.index[i],'TFBS']

        peaks_nearby_new = pd.concat([peaks_nearby_new, peaks_nearby_new_tmp], axis=0)
        peaks_nearby_new.index = np.arange(peaks_nearby_new.shape[0])
    
    peaks_nearby_new['pstart'] = peaks_nearby_new['TFBS'].str.split(split_symbol[0]).str[1].str.split(split_symbol[1]).str[0]
    peaks_nearby_new['pstart'] = peaks_nearby_new['pstart'].astype(float)
    peaks_nearby_new['pend'] = peaks_nearby_new['TFBS'].str.split(split_symbol[0]).str[1].str.split(split_symbol[1]).str[1]
    peaks_nearby_new['pend'] = peaks_nearby_new['pend'].astype(float)

    peaks_nearby_new['overlap'] = peaks_nearby_new.apply(lambda x: min(x['pend'],x['end']) - max(x['pstart'],x['start']), axis=1)

    df = peaks_nearby_new.copy()
    df = df.drop(['pstart', 'pend'], axis=1)

    df.columns = ['Gene','Strand','chr','start','end','Peak','overlap']
    df = df.sort_values(by='overlap', ascending=False)
    df = df.loc[~df.iloc[:,[0,5]].duplicated(),:].copy()
    
    #df = df.iloc[:,[0,5,7,9,8,10]].copy()
    df.index = np.arange(df.shape[0])
    
    return df



#=========================================================================================
# Homer related tools
#=========================================================================================
def PFM2Motif(file_name, out_file, detection_threshold=0):
    file = open(file_name, 'r').readlines()

    df_dict = {}
    for i in range(1,5):
        line = file[i].strip('\n').split(' ')
        line = list(filter(lambda a: a != '', line))
        line.remove('[')
        line.remove(']')
        #line = line[1:]
        df_dict[line[0]] = [float(i) for i in line[1:]]
    df = pd.DataFrame(df_dict)
    df = df[['A','C','G','T']]
    df[df.columns] = df.to_numpy() / df.sum(axis=1).to_numpy()[:, np.newaxis]

    motif_seq = ''.join(df.columns[np.argmax(df, axis=1)])
    motif_header = file[0]
    motif_header = motif_header.strip('>')
    motif_header = motif_header.strip('\n').split('\t')
    motif_header = '_'.join(motif_header)
    motif_header = '>' + motif_seq + '\t' + motif_header + '\t0'

    motif_file = open(out_file,'w')
    motif_file.write(motif_header)
    for i in range(df.shape[0]):
        freq_list = df.iloc[i,:].tolist()
        freq_list = [str(round(j, 6)) for j in freq_list]
        motif_file.write('\n')
        motif_file.write('\t'.join(freq_list))
        
    motif_file.close()




def peak2HomerInput(peaks, out_file, filetype='peaks', split_symbol=['-','-']):

    if filetype == 'peaks':
        homer_df = pd.DataFrame({'index': np.arange(len(peaks)),
                                 'chrom': [i.split(split_symbol[0],1)[0] for i in peaks],
                                 'start': [i.split(split_symbol[0],1)[1].split(split_symbol[1])[0] for i in peaks],
                                 'end': [i.split(split_symbol[0],1)[1].split(split_symbol[1])[1] for i in peaks],
                                 'strand':'.'})

    elif filetype == 'bed':
        homer_df = pd.DataFrame({'chrom': [i.split(split_symbol[0],1)[0] for i in peaks],
                                 'start': [i.split(split_symbol[0],1)[1].split(split_symbol[1])[0] for i in peaks],
                                 'end': [i.split(split_symbol[0],1)[1].split(split_symbol[1])[1] for i in peaks],
                                 'index': np.arange(len(peaks)),
                                 'useless': '.',
                                 'strand':'.'})
    
    homer_df.to_csv(out_file, index=False, header=False, sep='\t')




def run_HOMER_motif(peaks, out_dir, prefix, ref_genome, 
                    homer_path=None, split_symbol=['-','-'], size=200):

    """
    Function to run Homer from Python script.
    It will prepare Homer required input file
    and output results in the directory specified

    Arguments
    ---------
        peaks: list or array-like, peaks of interests

        out_dir: output directory to save the results
    
        prefix: prefix of all files, folder called out_dir/homer_prefix will be created to save all the results

        ref_genome: str, reference genome name, e.g., 'hg19', 'hg38'

        homer_path: path to Homer software, if Homer already added to the PATH, argument can be ignored

        split_symbol: how peak location ID is merged
                      'chr1-12345-23456' - split_symbol=['-','-']
                      'chr1:12345-23456' - split_symbol=[':','-']
        
        size: Homer paramter, the size of the region used for motif finding


    Returns
    ---------
        Homer results saved in out_dir/homer_prefix
        Peak list DataFrame in BED format

    """

    homer_df = pd.DataFrame({'index': np.arange(len(peaks)),
                             'chrom': [i.split(split_symbol[0],1)[0] for i in peaks],
                             'start': [i.split(split_symbol[0],1)[1].split(split_symbol[1])[0] for i in peaks],
                             'end': [i.split(split_symbol[0],1)[1].split(split_symbol[1])[1] for i in peaks],
                             'strand':'.'})
    
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)

    out_file = os.path.join(out_dir, '%s.txt'%(prefix))
    print("Save peaks list in BED file format at %s"%out_file)
    homer_df.to_csv(out_file, index=False, header=False, sep='\t')

    homer_dir = os.path.join(out_dir, 'homer_%s'%(prefix))
    if not os.path.exists(homer_dir):
        print("Creat HOMER output folder at %s"%homer_dir)
        os.mkdir(homer_dir)


    if homer_path is None:
        try:
            os.system('findMotifsGenome.pl %s %s %s -size %s'%(out_file, ref_genome, homer_dir, size))
        except:
            print("HOMER run failed. The HOMER peak format DataFrame is returned.")
            return homer_df
    else:
        homer_cmd_path = os.path.join(homer_path,'bin/findMotifsGenome.pl')
        try:
            os.system('%s %s %s %s -size %s'%(homer_cmd_path, out_file, ref_genome, homer_dir, size))
        except:
            print("HOMER run failed. The HOMER peak format DataFrame is returned.")
            return homer_df

    print("HOMER finished successfully! Please check the HTML report for interesting motifs.")
    print("motif_summary can be run with the motif index for further analysis.")
    return homer_df




def motif_summary(peak_file, homer_dir, motif_index, ref_genome,
                  homer_path=None, size=200):

    """
    Function to extract related peaks from motifs of interests.


    Arguments
    ---------
        peak_file: out_dir/prefix.peaks.bed file generated in run_HOMER_motif()

        homer_dir: out_dir/homer_prefix in run_HOMER_motif() | Homer output folder
    
        motif_index: motif of interests index in homer_dir/knownResults.html

        ref_genome: str, reference genome name, e.g., 'hg19', 'hg38'

        homer_path: path to Homer software, if Homer already added to the PATH, argument can be ignored
        
        size: Homer paramter, the size of the region used for motif finding
              Keep the same as in run_HOMER_motif()


    Returns
    ---------
        Motif related peaks information saved in homer_dir/
        DataFrame containing peak list and motif matching quality information

    """

    motif_file = os.path.join(homer_dir, 'knownResults/known%s.motif'%motif_index)
    motif_out = os.path.join(homer_dir, 'motif%s.peaks.txt'%motif_index)
    if homer_path is None:
        try:
            os.system('findMotifsGenome.pl %s %s %s -find %s -size %s > %s'%(
                peak_file, ref_genome, homer_dir, motif_file, size, motif_out))
        except:
            print("HOMER run failed. Make sure homer_path is set and all file paths are right.")
            return
    else:
        homer_cmd_path = os.path.join(homer_path,'bin/findMotifsGenome.pl')
        try:
            os.system('%s %s %s %s -find %s -size %s > %s'%(
                homer_cmd_path, peak_file, ref_genome, homer_dir, motif_file, size, motif_out))
        except:
            print("HOMER run failed. Make sure homer_path is set and all file paths are right.")
            return

    print("HOMER finished successfully! Motif related peaks will be loaded and returned.")
    motif_peaks = pd.read_csv(motif_out,sep='\t')
    peak_df = pd.read_csv(peak_file, sep='\t',header=None)

    peaks = peak_df.loc[motif_peaks['PositionID'].tolist()]
    peaks = peaks[0] + '-' + peaks[1].astype(str) + '-' + peaks[2].astype(str)
    motif_peaks.insert(0, 'peaks', peaks.to_numpy())

    return motif_peaks












import numpy as np
import pandas as pd
import scanpy as sc
import random
import anndata as ad
import time

from treasmo.moran_vec import Moran
import treasmo.lee_vec as lee_vec
from libpysal.weights import W

import scipy as sp
from scipy import stats
#from statsmodels.stats.multitest import multipletests
from statsmodels.stats.multitest import fdrcorrection

__author__ = "Chaozhong Liu"
__email__ = "czliubioinfo@gmail.com"



######
# add self-connection or not
# remove dropout: Default False
######



def Morans_I(mudata, mods=['rna','atac'], seed=1, max_RAM=16):
    """ 
    Function to calculate Moran's I for all the features in anndat_multiome

    Arguments
    --------------

        mudata: single-cell multi-omics data saved as MuData object

        mods: scRNA-seq and scATAC-seq modality name in MuData object

        seed: random seed to make the results reproducible

        max_RAM: maximum limitation of memory (Gb)


    Returns
    --------------
        mudata added with
            var['Morans.I']

    """

    # Check if mods exist in MuData
    if sum([i in mudata.mod.keys() for i in mods]) == len(mods):
        pass
    else:
        print("Not all mods provided are in MuData! Running stopped.")
        return

    # Start
    random.seed(seed)
    np.random.seed(seed)
    start = time.time()

    # Construct connection matrix
    cong_mtx = mudata.obsp['connectivities'].toarray()
    #cong_mtx = cong_mtx/cong_mtx.sum(axis=1)[:,np.newaxis]
    neighbors = {}
    weights = {}

    for i in range(cong_mtx.shape[0]):
        neighbors[i] = np.nonzero(cong_mtx[i])[0].tolist()
        weights[i] = cong_mtx[i][np.nonzero(cong_mtx[i])[0]].tolist()
    w = W(neighbors, weights)

    mi = Moran(mudata.n_obs, w, permutations=0)
    mudata.var['Morans.I'] = np.NaN

    for mod in mods:
        mi.calc_i(mudata.mod[mod].X, seed=seed, max_RAM=max_RAM)
        mudata.mod[mod].var['Morans.I'] = mi.I

    mudata.var['Morans.I'] = np.concatenate([mudata.mod[mod].var['Morans.I'].to_numpy() for mod in mudata.mod.keys()])
    
    #print(mi.I)
    print("Finished calculating Moran's I. %.3fs past"%(time.time()-start))
    return mudata




def _check_array(mtx):
    if sp.sparse.issparse(mtx): #isinstance(mtx, sp.spmatrix):
        return mtx.toarray()
    elif isinstance(mtx, np.ndarray):
        return mtx
    else:
        raise Exception("Omics Data should be either numpy array or scipy sparse matrix.")




def _prepare_mtx(anndat, features):

    index_df = pd.DataFrame({'index':np.arange(anndat.n_vars)})
    index_df.index = anndat.var.index

    feature_index = index_df.loc[features,'index'].to_numpy()

    return _check_array(anndat.X[:,feature_index])




def Global_L(mudata, pairsDf, mods=['rna','atac'], permutations=0, percent=0.1, seed=1, max_RAM=16):
    """ 
    Function to calculate the global L index for all the pairs in anndat_multiome

    Arguments
    --------------

        mudata: single-cell multi-omics data saved as MuData object

        mods: scRNA-seq and scATAC-seq modality name in MuData object

        pairsDf: gene-peak pair DataFrame containing pairs to be calculated
                 self-prepared or call treasmo.tl.peaks_within_distance / TFBS_match

        permutations: number of permutations for significance test.
                      Default is 0, meaning no significance test
                      999 is a good choice for the test, but it might take a long time (hours) to finish depending on the number of pairs
        
        percent: percentage of cells to shuffle during permutation.
                 For most of the time, default 0.1 is already a good choice.
        
        seed: random seed to make the results reproducible

        max_RAM: maximum limitation of memory (Gb)


    Returns
    --------------
        pairsDf added with extra columns:
            Global L results
            QC metrics (sparsity)

    """

    random.seed(seed)
    np.random.seed(seed)
    start = time.time()

    print("Calculating KNN graph-based global correlation...")
    cong_mtx = mudata.obsp['connectivities'].toarray()
    np.fill_diagonal(cong_mtx, 1.0)
    eSP = lee_vec.Spatial_Pearson(cong_mtx, permutations=permutations)

    genes = pairsDf['Gene']
    peaks = pairsDf['Peak']

    gene_X = _prepare_mtx(mudata.mod[mods[0]], genes)
    peak_X = _prepare_mtx(mudata.mod[mods[1]], peaks)

    eSP = eSP.fit(gene_X, peak_X, percent=percent, seed=seed, max_RAM=max_RAM)

    pairsDf['L'] = eSP.association_
    pairsDf['L.p_value'] = eSP.significance_ if permutations else np.NaN

    if permutations:
        _,pairsDf['L.FDR'] = fdrcorrection(peaks_nearby['L.p_value'],
                                           alpha=0.05, method='indep')
    else:
        pairsDf['L.FDR'] = np.NaN

    pairsDf['gene.pct'] = mudata.mod[mods[0]].var.loc[genes,'Frac.all'].to_numpy()
    pairsDf['peak.pct'] = mudata.mod[mods[1]].var.loc[peaks,'Frac.all'].to_numpy()

    print("Finished calculating correlation. %.3fs past"%(time.time()-start))

    return pairsDf




def _calculate_LL(mudata, genes, peaks,
                  mods=['rna','atac'],
                  seed=1, max_RAM=16):
    random.seed(seed)
    np.random.seed(seed)

    cong_mtx = mudata.obsp['connectivities'].toarray()
    np.fill_diagonal(cong_mtx, 1.0)

    gene_X = _prepare_mtx(mudata.mod[mods[0]], genes)
    peak_X = _prepare_mtx(mudata.mod[mods[1]], peaks)

    eSP2 = lee_vec.Spatial_Pearson_Local(cong_mtx, permutations=0)
    eSP2 = eSP2.fit(gene_X,peak_X,seed=seed, max_RAM=max_RAM)

    '''
    if permutations:
        local_p_df = pd.DataFrame(eSP2.significance_)
        local_p_df.columns = pair_names
    '''

    L_mtx = eSP2.associations_
    pair_names = pd.Series(genes) + '~' + pd.Series(peaks)
    pair_names = pair_names.to_numpy()

    return L_mtx, pair_names




def Local_L(mudata, genes, peaks,
            mods=['rna','atac'],
            rm_dropout=False,
            seed=1, max_RAM=16):
    """ 
    Function to calculate the local L index for all the pairs in anndat_multiome

    Arguments
    --------------

        mudata: single-cell multi-omics data saved as MuData object

        mods: scRNA-seq and scATAC-seq modality name in MuData object

        genes, peaks: one-to-one lists containing gene and peak names

        rm_dropout: make local L index to be 0 if feature value is 0 (dropout)

        seed: random seed to make the results reproducible

        max_RAM: maximum limitation of memory (Gb)


    Returns
    --------------
        mudata added with
            Local L matrix and gene-peak pair names

    """

    random.seed(seed)
    np.random.seed(seed)
    start = time.time()

    print("Inferring KNN graph-based local correlation...")
    
    L_mtx, L_mtx_names = _calculate_LL(mudata, genes, peaks,
                                       mods=mods,
                                       seed=1, max_RAM=16)

    if rm_dropout:
        print("Setting local correlation to 0 for cells with no expression on either feature of a certain pair...")
        GP_names = L_mtx_names
        Dropout_mtx = _dropout_filter(mudata, GP_names, mods)
        L_mtx = L_mtx * Dropout_mtx

    mudata.uns['Local_L'] = L_mtx
    mudata.uns['Local_L_names'] = L_mtx_names

    print("Following changes made to the AnnData object:")
    print("\tKNN graph-based local correlation results saved in uns['Local_L']")
    print("\tGene-peak pair names saved in uns['Local_L_names']")

    print("Finished Inferring local correlation. %.3fs past"%(time.time()-start))
    return mudata




def _dropout_filter(mudata, GP_names, mods):

    # Gene dropout
    index_df = pd.DataFrame({'index':np.arange(len(mudata.mod[mods[0]].var_names))})
    index_df.index = mudata.mod[mods[0]].var_names
    GP_G = [gp.split('~')[0] for gp in GP_names]
    GP_genes_index = index_df.loc[GP_G,:]['index'].to_numpy()

    E_mtx = _check_array(mudata.mod[mods[0]].X)
    E_mtx_dropout_value = np.zeros(E_mtx.shape)
    Dropout_mtx = (~np.isclose(E_mtx_dropout_value,E_mtx, 1e-3)).astype(int)
    Dropout_mtx_G = Dropout_mtx[:,GP_genes_index]

    # Peak dropout
    index_df = pd.DataFrame({'index':np.arange(len(mudata.mod[mods[1]].var_names))})
    index_df.index = mudata.mod[mods[1]].var_names
    GP_P = [gp.split('~')[1] for gp in GP_names]
    GP_peaks_index = index_df.loc[GP_P,:]['index'].to_numpy()

    E_mtx = _check_array(mudata.mod[mods[1]].X)
    E_mtx_dropout_value = np.zeros(E_mtx.shape)
    Dropout_mtx = (~np.isclose(E_mtx_dropout_value,E_mtx, 1e-3)).astype(int)
    Dropout_mtx_P = Dropout_mtx[:,GP_peaks_index]
    
    Dropout_mtx = Dropout_mtx_G * Dropout_mtx_P
    
    return Dropout_mtx





#=========================================================================================
# Pearson correlation function for benchmark purpose
#=========================================================================================

def Pearsonr(mudata, genes, peaks,
                   mods=['rna','atac'],
                   p_value=False, seed=1):
    random.seed(seed)
    np.random.seed(seed)

    gene_X = _prepare_mtx(mudata.mod[mods[0]], genes)
    peak_X = _prepare_mtx(mudata.mod[mods[1]], peaks)

    gene_X = pd.DataFrame(gene_X, index=np.arange(gene_X.shape[0]), columns=np.arange(gene_X.shape[1]))
    peak_X = pd.DataFrame(peak_X, index=np.arange(peak_X.shape[0]), columns=np.arange(gene_X.shape[1]))

    global_P_df = pd.DataFrame(peak_X.corrwith(gene_X, method='pearson', axis=0))
    global_P_df.index = pd.Series(genes) + '~' + pd.Series(peaks)
    global_P_df.columns = ['r']

    if p_value:
        print("Calculating p-values...")
        p_list = []
        for i in range(global_P_df.shape[0]):
            p_list.append(stats.pearsonr(gene_X.iloc[:,i].to_numpy(), peak_X.iloc[:,i].to_numpy())[1])
        global_P_df['r.p_value'] = p_list
        print("Calculating FDR...")
        _, global_P_df['r.FDR'] = fdrcorrection(global_P_df['r.p_value'],
                                                alpha=0.05, method='indep')
    else:
        global_P_df['r.p_value'] = np.NaN
        global_P_df['r.FDR'] = np.NaN

    return global_P_df










import numpy as np
import pandas as pd
import anndata as ad
import time
import math
import scipy as sp

import seaborn as sns
import scanpy as sc
import matplotlib.pyplot as plt
from matplotlib.pyplot import figure
from matplotlib.lines import Line2D
import matplotlib.colors as mcolors


from scipy.cluster.hierarchy import dendrogram
from sklearn.cluster import AgglomerativeClustering
from sklearn.gaussian_process import GaussianProcessRegressor

import os
import sys
from sklearn import preprocessing


__author__ = "Chaozhong Liu"
__email__ = "czliubioinfo@gmail.com"


#=========================================================================================
# Local correlation heatmap
# Gene-peak correlation Feature plot
# Tractory-based regulatory dynamics
# Pairs Spearman correlation heatmap (demonstrate the relationship between regulatory pairs)
#=========================================================================================



# Heatmap
def LocalCor_Heatmap(mudata, pairs, groupby, cluster=True, save=None, **kwds):
    """
    Function to visualize the local L matrix by heatmap, and cluster features

    Arguments
    ---------
        mudata: single-cell multi-omics data saved as MuData object

        pairs: gene-peak pair lists to visualize

        cluster: cluster features or not

        groupby: if provided, cells / pseudo-bulks will be grouped by the label name

        save: if provided, heatmap will be saved in the file path

        **kwds: other arguments for sc.pl.heatmap()


    Returns
    ---------
        Local L index heatmap with features clustered or not


    """

    L_mtx = mudata.uns['Local_L']
    L_mtx_name = mudata.uns['Local_L_names']

    if not np.all(np.isin(np.asarray(pairs), L_mtx_name)):
        print("The following pairs listed are in the local correlation matrix:")
        print(", ".join(np.asarray(pairs)[~np.isin(np.asarray(pairs), L_mtx_name)]))
        pairs = np.asarray(pairs)[np.isin(np.asarray(pairs), L_mtx_name)]

    index_df = pd.DataFrame({'index':np.arange(L_mtx_name.shape[0])})
    index_df.index = L_mtx_name
    idx = index_df.loc[pairs,'index'].to_numpy()

    anndat_L = ad.AnnData(
        X = L_mtx[:,idx],
        obs = mudata.obs
    )
    anndat_L.var_names = pairs
    anndat_L.var_names_make_unique()

    if f'{groupby}_colors' in mudata.uns.keys():
        anndat_L.uns[f'{groupby}_colors'] = mudata.uns[f'{groupby}_colors']

    if cluster:
        _Heatmap_cluster(anndat_L, groupby=groupby, save=save, **kwds)
    else:
        _Heatmap_nocluster(anndat_L, groupby=groupby, save=save, **kwds)




def _Heatmap_cluster(anndat_L, groupby, save=None, **kwds):

    # Determine features order in heatmap
    model = AgglomerativeClustering(n_clusters=15, affinity='euclidean', 
                                  linkage='ward', compute_distances=True)
    model = model.fit(anndat_L.X.T)
    cluster_labels = model.labels_
    counts = np.zeros(model.children_.shape[0])
    n_samples = len(model.labels_)
    for i, merge in enumerate(model.children_):
        current_count = 0
        for child_idx in merge:
            if child_idx < n_samples:
                current_count += 1  # leaf node
            else:
                current_count += counts[child_idx - n_samples]
        counts[i] = current_count

    linkage_matrix = np.column_stack(
        [model.children_, model.distances_, counts]
    ).astype(float)
    R = dendrogram(linkage_matrix, no_plot=True)
    feature_order = R['leaves']

    anndat_L = anndat_L[:,feature_order]
    if save is None:
        sc.pl.heatmap(anndat_L, anndat_L.var_names,
                      groupby=groupby, **kwds)
                      #vmin=-1.0, vmax=1.0, dendrogram=True, cmap='vlag', figsize=(10,6)
    else:
        with plt.rc_context():  # Use this to set figure params like size and dpi
            sc.pl.heatmap(anndat_L, anndat_L.var_names,
                          show=False,
                          groupby=groupby, **kwds)
            plt.savefig(save)



def _Heatmap_nocluster(anndat_L, groupby, save=None, **kwds):

    if save is None:
        sc.pl.heatmap(anndat_L, anndat_L.var_names,
                      groupby=groupby, **kwds)
                      #vmin=-1.0, vmax=1.0, dendrogram=True, cmap='vlag', figsize=(10,6)
    else:
        with plt.rc_context():  # Use this to set figure params like size and dpi
            sc.pl.heatmap(anndat_L, anndat_L.var_names,
                          show=False,
                          groupby=groupby, **kwds)
            plt.savefig(save)


def _check_array(mtx):
    if sp.sparse.issparse(mtx): #isinstance(mtx, sp.spmatrix):
        return mtx.toarray()
    elif isinstance(mtx, np.ndarray):
        return mtx
    else:
        raise Exception("Omics Data should be either numpy array or scipy sparse matrix.")



# Feature Plot in UMAP
def visualize_marker(mudata, gene, peak, mods=['rna','atac'], 
                     cmaps='plasma', basis='umap',
                     vmins=None, vmaxs=None,
                     figsize=None, save=None, **kwds):

    """
    Function to visualize the gene-peak pair correlation in UMAP

    Arguments
    ---------
        mudata: single-cell multi-omics data saved as MuData object

        gene: gene name

        peak: peak name

        mods: scRNA-seq and scATAC-seq modality name in MuData object

        cmaps: Color map to use for continous variables. Could be either a single color_map or a list

        basis: the embeddings to plot. Could be either a single embedding space or a list for each of the feature

        vmins: min value to color. Could be either a single value or a list for gene, peak, and correlation

        vmaxs: max value to color. Same as vmins

        figsize: figure size

        save: str. if provided, heatmap will be saved in the file path

        **kwds: other arguments for sc.pl.embedding


    Returns
    ---------
        Embedding colored by the gene, peak, and the correlation between gene and peak


    """

    if isinstance(basis, str):
        basis = [basis, basis, basis]

    if isinstance(cmaps, str):
        cmaps = [cmaps, cmaps, cmaps]

    
    _, n_col = mudata.obs.shape

    anndat_sp_L = pd.DataFrame(mudata.uns['Local_L'])
    anndat_sp_L.columns = mudata.uns['Local_L_names']

    feature = f'{gene}~{peak}'
    mudata.obs[feature] = anndat_sp_L[feature].to_numpy()
    mudata.obs[gene] = _check_array(mudata.mod[mods[0]].X[:,mudata.mod[mods[0]].var_names == gene]).squeeze() #.toarray()
    mudata.obs[peak] = _check_array(mudata.mod[mods[1]].X[:,mudata.mod[mods[1]].var_names == peak]).squeeze() #.toarray()

    if vmins is None:
        vmins = [mudata.obs[gene].min(), mudata.obs[peak].min(), mudata.obs[feature].min()]

    if vmaxs is None:
        vmaxs = [mudata.obs[gene].max(), mudata.obs[peak].max(), mudata.obs[feature].max()]

    print('%s and %s'%(gene, peak))

    if save is None:
        _, axs = plt.subplots(ncols=3, figsize=figsize)

        for i in range(3):
            sc.pl.embedding(mudata, basis=basis[i],
                            ax=axs[i],
                            color=[gene, peak, feature][i],
                            cmap=cmaps[i],
                            vmin=vmins[i], vmax=vmaxs[i],
                            show=False,
                            **kwds)

        plt.show()

    else:
        with plt.rc_context():  # Use this to set figure params like size and dpi
            _, axs = plt.subplots(ncols=3, figsize=figsize)

            for i in range(3):
                sc.pl.embedding(mudata, basis=basis[i],
                                ax=axs[i],
                                color=[gene, peak, feature][i],
                                cmap=cmaps[i],
                                vmin=vmins[i], vmax=vmaxs[i],
                                show=False,
                                **kwds)
            plt.savefig(save)

    mudata.obs = mudata.obs.iloc[:,0:n_col]




# Visualize trajectory regulatory dynamic for a given gene
def PathDynamics(mudata, ident, path, gene, peaks=None,
                 xlim=None, ylim=None,
                 title=None, title_fontsize=15, ticks_fontsize=12,
                 x_label='Pseudotime', y_label='Correlation Strength', label_fontsize=12,
                 curve_colors=None, dot_size=5, linewidth=3,
                 ident_colors=None, show_legend=True, save=None):

    """
    Function to visualize the gene-peak pair correlation changes along pseudotime + cell type proportion visualization

    Arguments
    ---------
        mudata: single-cell multi-omics data saved as MuData object

        ident: ident label in mudata.obs columns to distinguish clusters listed in path

        path: list of clusters ordered by their sequence on the trajectory. A path here should have no branch.

        gene: gene name

        peaks: list of peak names to be paired with the gene

        *To visualize the results, need to run df.PathDynamics() first*

        xlim: (min, max) the pseudotime range limit

        ylim: (min, max) the correlation range limit, useful to remove outliers

        (title/ticks/label)_fontsize: fontsize of different element

        (x/y)_label: labels for x/y axis

        curve_colors: curve colors for each of the gene-peak pair correlation; if not specified, defaul color palette will be applied.

        dot_size: dot size in plot

        linewidth: curve width

        ident_colors: colors of each cluster to be plotted in the proportion bar. If not specified, function will look for uns[IDENT_colors] first. If not found, default color palette will be applied.

        show_legend: Show color legend or not

        save: str. if provided, heatmap will be saved in the file path


    Returns
    ---------
        Embedding colored by the gene, peak, and the correlation between gene and peak


    """

    path_name = '_'.join(path)

    try:
        data = mudata.uns['pathDym'][path_name][gene].copy()
    except:
        raise Exception(f"Path regulatory dynamics data not found in uns['pathDym']['{path_name}'']['{gene}'']. Please run ds.PathDynamics() first.")


    # Filter by xlim
    if xlim is not None:
        n_before = data.shape[0]
        data = data.loc[(data['time']>=xlim[0]) & (data['time']<=xlim[1]), :]
        n_after = data.shape[0]
        if n_after < n_before:
            print(f"{n_before-n_after} time bins removed from plotting")


    # Get the features name
    time_index = data.columns.tolist().index('time')
    if peaks is None:
        features = data.columns.tolist()[0:time_index]
    else:
        features = [f"{gene}~{peak}" for peak in peaks]
        data = data.loc[:, features+['time']+path]


    # Calculate cumulative cell proportion
    prpmtx = data[path].to_numpy()
    prpmtx = (prpmtx / prpmtx.sum(axis=1)[:, np.newaxis])
    data[path] = prpmtx
    data[path] = data[path].cumsum(axis=1)
    data[path[-1]] = 1.0

    # Select colors for features and clusters
    color_tmp = plt.cm.rainbow(np.linspace(0, 1, len(features)+len(path)))
    # curve color
    if curve_colors is None:
        curve_colors = []
        for i in range(len(features)):
            curve_colors.append(mcolors.to_hex(color_tmp[i]))
        #curve_colors = np.array(sc.pl.palettes.zeileis_28)[0:len(features)]
    else:
        if len(curve_colors) == len(features):
            pass
        else:
            raise Exception(f"Number of curve colors {len(curve_colors)} doesn't match number of features {len(features)}")
    # cluster color
    if ident_colors is None:
        if f'{ident}_colors' in mudata.uns.keys():
            ident_colors = list(mudata.uns[f'{ident}_colors'])
            path_index = [mudata.obs[ident].cat.categories.tolist().index(cluster) for cluster in path]
            ident_colors = [ident_colors[i] for i in path_index]
        else:
            print(f"{ident}_colors not found in mudata.uns, will use defaul color palette.")
            ident_colors = []
            for i in range(len(features), len(features)+len(path)):
                ident_colors.append(mcolors.to_hex(color_tmp[i]))
            #ident_colors = np.array(sc.pl.palettes.zeileis_28)[-len(path):]

    # Prepare for plot
    if title is None:
        title = f"{'-'.join(path)}: {gene}"

    curveDf = data.loc[:, features+['time']].copy()
    prpDf = data.loc[:, path+['time']].copy()
    prpDf.columns.name = ident

    # Set up the figure
    fig1 = plt.figure(constrained_layout=False)
    gs1 = fig1.add_gridspec(nrows=10, ncols=10, left=0.0, right=1.0, top=0.9, bottom=0.3, wspace=0.05, hspace=0.2)
    f1_ax1 = fig1.add_subplot(gs1[:-1, :-1])
    f1_ax2 = fig1.add_subplot(gs1[-1, :-1])
    f1_ax3 = fig1.add_subplot(gs1[:-1, -1])
    hfont = {'fontname':'sans'}
    fig1.suptitle(title, fontsize=title_fontsize, x=0.45 , y=0.95, verticalalignment='bottom', **hfont)

    _ComplexCurvePlot(curveDf, prpDf, f1_ax1, f1_ax2, xlim=xlim, ylim=ylim,
                      ticks_fontsize=ticks_fontsize,
                      x_label=x_label, y_label=y_label, label_fontsize=label_fontsize,
                      curve_colors=curve_colors, dot_size=dot_size, linewidth=linewidth,
                      ident_colors=ident_colors)

    # Legend
    if show_legend:
        legend_elements = [Line2D([0], [0], linewidth=0, label='Curves')]
        legend_elements += [Line2D([0], [0], linewidth=0, marker='o', color=curve_colors[i], label=features[i].split('~')[1], markersize=7) for i in range(len(features))]
        legend_elements += [Line2D([0], [0], linewidth=0, label='')]
        legend_elements += [Line2D([0], [0], linewidth=0, label=ident)]
        legend_elements += [Line2D([0], [0], linewidth=0, marker='o', color=ident_colors[i], label=path[i], markersize=7) for i in range(len(path))]
        #ax.set_title(feature)
        f1_ax3.legend(handles=legend_elements, loc='center left', bbox_to_anchor=(0.1, 0.5), frameon=False)
        f1_ax3.tick_params(left = False, right = False , labelleft = False ,
                           labelbottom = False, bottom = False)
        f1_ax3.grid(which='both', linestyle='-', linewidth='0')
        f1_ax3.axis('off')

    if save is None:
        plt.show()
    else:
        with plt.rc_context():
            plt.savefig(save)




def _ComplexCurvePlot(curveDf, prpDf, f1_ax1, f1_ax2, xlim=None, ylim=None,
                      ticks_fontsize=12,
                      x_label='Pseudotime', y_label='Correlation Strength', label_fontsize=12,
                      curve_colors=None, dot_size=5, linewidth=3,
                      ident_colors=None):
    
    # Prepare some arguments
    if isinstance(curveDf, dict):
        features = list(curveDf.keys())
    else:
        features = curveDf.columns.tolist()
        features.remove('time')
    if isinstance(linewidth, int):
        linewidth = [linewidth] * len(features)
    path = prpDf.columns.tolist()
    path.remove('time')
    ident = prpDf.columns.name


    # Plot the curves
    if isinstance(curveDf, dict):
        time_range = (curveDf[list(curveDf.keys())[0]]['time'].min(), curveDf[list(curveDf.keys())[0]]['time'].max())
        i_key = 0
        for key_name, data_tmp in curveDf.items():
            _curve_plot(data_tmp, feature=key_name, colors=curve_colors[i_key], xlim=xlim, ylim=ylim, ax=f1_ax1, dot_size=dot_size, linewidth=linewidth[i_key])
            i_key += 1
    else:
        time_range = (curveDf['time'].min(), curveDf['time'].max())
        for i in range(len(features)):
            data_tmp = curveDf[['time',features[i]]].copy()
            data_tmp.columns = ['time', 'value']
            _curve_plot(data_tmp, feature=features[i], colors=curve_colors[i], xlim=xlim, ylim=ylim, ax=f1_ax1, dot_size=dot_size, linewidth=linewidth[i])
    
    if xlim is None:
        f1_ax1.set_xlim(time_range[0], time_range[1])
    else:
        f1_ax1.set_xlim(xlim[0], xlim[1]) #data['time'].min(), data['time'].max())
    if ylim is not None:
        f1_ax1.set_ylim(ylim[0], ylim[1]) 
    f1_ax1.tick_params(left = True, right = False , labelleft = True ,
                       labelbottom = False, bottom = False)
    f1_ax1.grid(which='both', linestyle='-', linewidth='0')
    f1_ax1.set_ylabel(y_label, fontsize=label_fontsize)
    #Font size in plots
    f1_ax1.yaxis.label.set_fontsize(label_fontsize)
    f1_ax1.tick_params(axis='both', which='major', labelsize=ticks_fontsize)


    # Plot the proportion bar
    if f1_ax2 is None:
        pass
    else:
        _prp_var(prpDf, path, colors=ident_colors, xlim=xlim, ax=f1_ax2)
        f1_ax2.set_xlabel(x_label, fontsize=label_fontsize)
        #Font size in plots
        f1_ax2.xaxis.label.set_fontsize(label_fontsize)
        f1_ax2.tick_params(axis='both', which='major', labelsize=ticks_fontsize)




def _curve_plot(data, feature, colors, dot_size, linewidth, xlim=None, ylim=None, ax=None):

    if ax is None:
        ax = plt.gca()

    #for i in range(len(features)):
    # define the model and draw some data
    xdata = data['time'].to_numpy()
    ydata = data['value'].to_numpy()

    #filter by ylim
    #if ylim is not None:
    #    filter_bool = (ydata >= ylim[0]) & (ydata <= ylim[1])
        #xdata = xdata[filter_bool]
        #ydata = ydata[filter_bool]
        #if np.sum(~filter_bool) > 0:
        #    print(f"{feature}: {np.sum(~filter_bool)} time bins removed based on provided ylim.")
    
    filter_na = ~np.isnan(ydata)
    xdata = xdata[filter_na]
    ydata = ydata[filter_na]

    # Compute the Gaussian process fit
    gp = GaussianProcessRegressor(random_state=1)
    gp.fit(xdata[:, np.newaxis], ydata)
    xfit = np.linspace(data['time'].min(), data['time'].max(), 1000)
    yfit, std = gp.predict(xfit[:, np.newaxis], return_std=True)
    dyfit = 1.96 * std #np.sqrt(MSE)
    
    # confidence interval
    '''
    stdDf = data[['time','value']].groupby('time').std()
    stdata = stdDf['value'].to_numpy()
    timePoint = stdDf.index.to_numpy()

    gp = GaussianProcessRegressor(random_state=1)
    gp.fit(timePoint[:, np.newaxis], timePoint)
    stdfit, _ = gp.predict(xfit[:, np.newaxis], return_std=True)

    # Generate samples from posterior distribution. 
    y_hat_samples = gp.sample_y(xfit[:, np.newaxis], n_samples=1000)
    # Compute the mean of the sample. 
    y_hat = np.apply_over_axes(func=np.mean, a=y_hat_samples, axes=1).squeeze()
    # Compute the standard deviation of the sample. 
    stdfit = np.apply_over_axes(func=np.std, a=y_hat_samples, axes=1).squeeze()
    '''
    
    #ax.plot(xdata, ydata, 'or', size=dot_size)
    ax.scatter(xdata, ydata, s=dot_size, c=colors)
    ax.plot(xfit, yfit, '-', color=colors, linewidth=linewidth, alpha=0.8)
    ax.fill_between(xfit, yfit-dyfit, yfit+dyfit,
                    color=colors, alpha=0.5)




def _prp_var(data, path, colors, xlim=None, ax=None):

    if ax is None:
        ax = plt.gca()

    for i in range(len(path)):
        ax.plot(data['time'], data[path[i]], '-', linewidth=0)
        if i == 0:
            ax.fill_between(data['time'], 0, data[path[i]],
                             color=colors[i], alpha=0.75)
        else:
            ax.fill_between(data['time'], data[path[i-1]], data[path[i]],
                             color=colors[i], alpha=0.75)
    
    
    if xlim is None:
        ax.set_xlim(data['time'].min(), data['time'].max())
    else:
        ax.set_xlim(xlim[0], xlim[1])

    ax.set_ylim(0, 1)
    #ax.tight_layout()
    ax.tick_params(left = False, right = False , labelleft = False ,
                    labelbottom = True, bottom = True)
    ax.grid(which='both', linestyle='-', linewidth='0')



def DynamicSumMtx(mudata, ident, path, gene, peaks=None, 
                  feature_colors=None, show_legend=True, save=None, **kwds):

    path_name = '_'.join(path)

    try:
        data = mudata.uns['pathDym'][path_name][gene].copy()
    except:
        raise Exception(f"Path regulatory dynamics data not found in uns['pathDym']['{path_name}'']['{gene}'']. Please run ds.PathDynamics() first.")

    time_index = data.columns.tolist().index('time')
    #path = data.columns.tolist()[time_index+1:]
    if peaks is None:
        features = data.columns.tolist()[0:time_index]
    else:
        features = [f"{gene}~{peak}" for peak in peaks]

    if len(features) <= 1:
        raise Exception("Please select at least two gene-peak pairs for analysis.")

    corrDf = data[features].corr(method='spearman')
    corrDf.index = corrDf.index.str.split('~').str[1]
    corrDf.columns = corrDf.columns.str.split('~').str[1]

    # plot the heatmap
    mask = np.triu(np.ones_like(corrDf))
    np.fill_diagonal(mask, 0.0)

    fig, ax = plt.subplots()
    sns.heatmap(corrDf, mask=mask, **kwds)
    plt.grid(which='both', linestyle='-', linewidth='0')
    plt.xticks(rotation=0, fontsize=10)
    plt.yticks(rotation=0, fontsize=10)

    color_tmp = plt.cm.rainbow(np.linspace(0, 1, len(features)+len(path)))
    if feature_colors is None:
        feature_colors = []
        for i in range(len(features)):
            feature_colors.append(mcolors.to_hex(color_tmp[i]))
    else:
        if len(feature_colors) == len(features):
            pass
        else:
            raise Exception(f"Number of feature colors {len(feature_colors)} doesn't match number of features {len(features)}")

    symbolsx = ["\u26AB"] * len(features)
    colorsx = feature_colors
    ax.set_xticks([i+0.5 for i in range(len(symbolsx))])
    ax.set_xticklabels(symbolsx, size=25)
    for tick, color in zip(ax.get_xticklabels(), colorsx):
        tick.set_color(color)

    ax.set_yticks([i+0.5 for i in range(len(symbolsx))])
    ax.set_yticklabels(symbolsx, size=25)
    for tick, color in zip(ax.get_yticklabels(), colorsx):
        tick.set_color(color)

    if show_legend:
        legend_elements = [Line2D([0], [0], linewidth=0, label='Peaks')]
        legend_elements += [Line2D([0], [0], linewidth=0, marker='o', color=feature_colors[i], label=features[i].split('~')[1], markersize=7) for i in range(len(features))]
        ax.legend(handles=legend_elements, loc='center left', bbox_to_anchor=(1.4, 0.5), frameon=False)


    if save is None:
        plt.show()
    else:
        with plt.rc_context():
            plt.savefig(save)



def DynamicModule(mudata, somDict, prpDfin, xlim=None, ylim=None,
                  split=False, n_cols=3,
                  title=None, title_fontsize=15, ticks_fontsize=12,
                  x_label='Pseudotime', y_label='Correlation Strength', label_fontsize=12,
                  curve_colors=None, dot_size=5, linewidth=3,
                  ident_colors=None, show_legend=True, save=None):
    
    # Some variables
    modules = list(somDict.keys())
    prpDf = prpDfin.copy()
    ident = prpDf.columns.name
    path = prpDf.columns[:-1].tolist()


    # Select colors for modules and clusters
    color_tmp = plt.cm.rainbow(np.linspace(0, 1, len(modules)+len(path)))
    # curve color
    if curve_colors is None:
        curve_colors = []
        for i in range(len(modules)):
            curve_colors.append(mcolors.to_hex(color_tmp[i]))
    else:
        if len(curve_colors) == len(modules):
            pass
        else:
            raise Exception(f"Number of curve colors {len(curve_colors)} doesn't match number of modules {len(modules)}")
    # cluster color
    if ident_colors is None:
        if f'{ident}_colors' in mudata.uns.keys():
            ident_colors = list(mudata.uns[f'{ident}_colors'])
            path_index = [mudata.obs[ident].cat.categories.tolist().index(cluster) for cluster in path]
            ident_colors = [ident_colors[i] for i in path_index]
        else:
            print(f"{ident}_colors not found in mudata.uns, will use defaul color palette.")
            ident_colors = []
            for i in range(len(modules), len(modules)+len(path)):
                ident_colors.append(mcolors.to_hex(color_tmp[i]))


    # Prepare module-wise data for plotting
    dataDict = {}
    for sommod in somDict.keys():
        data = somDict[sommod].copy()

        # filter by xlim
        if xlim is not None:
            data = data.loc[(data['time']>=xlim[0]) & (data['time']<=xlim[1]), :]

        # wide to long
        long_list = [ pd.DataFrame(data[['time', feature]].to_numpy(), columns=['time','value']) for feature in data.columns[:-1]]
        for i in range(len(data.columns[:-1])):
            long_list[i].insert(2, 'feature', data.columns[:-1][i])
        longDf = pd.concat(long_list)
        longDf['module'] = sommod
        longDf.index = np.arange(longDf.shape[0])
        dataDict[sommod] = longDf

    # Prepare cell type proportion
    prpmtx = prpDf[path].to_numpy()
    prpmtx = (prpmtx / prpmtx.sum(axis=1)[:, np.newaxis])
    prpDf[path] = prpmtx
    prpDf[path] = prpDf[path].cumsum(axis=1)
    prpDf[path[-1]] = 1.0


    # Prepare for plot
    if title is None:
        title = f"{'-'.join(path)} Modules"


    if split:
        n_modules = len(modules)
        n_rows = math.ceil(n_modules/n_cols)

        fig1 = plt.figure(constrained_layout=False)
        gs = fig1.add_gridspec(n_rows, n_cols+1, #9*n_cols+1, 
                               left=0.0, right=(0.1+0.9*n_cols), 
                               bottom=0.0, top=(0.6*n_rows),
                               width_ratios = [9]*n_cols + [1],
                               hspace=0.15, wspace=0.2)
        #plt.subplots_adjust(wspace=None, hspace=None)

        ax_list = []
        max_col = 0
        max_row = {ir:0 for ir in range(n_cols)}
        for i in range(n_modules):
            index_col = i % n_cols
            if max_col < index_col:
                max_col = index_col

            index_row = i // n_cols
            if max_row[index_col] < index_row:
                max_row[index_col] = index_row
            
            gsij = gs[index_row:(index_row+1), index_col:(index_col+1)].subgridspec(10, 1, hspace=0.02)

            ax_list.append((
                            fig1.add_subplot(gsij[:-1,:]), 
                            fig1.add_subplot(gsij[-1,:])
                          ))

        f1_legend = fig1.add_subplot(gs[0:(index_row+1), (max_col+1):(max_col+2)])

        hfont = {'fontname':'sans'}
        fig1.suptitle(title, fontsize=title_fontsize, x=(0.1+0.9*n_cols)*(9*n_cols)/(9*n_cols+1)/2 , y=(0.6*n_rows)+0.05, verticalalignment='bottom' , **hfont)

        for i in range(len(ax_list)):
            index_col = i % n_cols
            index_row = i // n_cols
            
            tmp_ax1, tmp_ax2 = ax_list[i]
            data = somDict[modules[i]].copy()
            # filter by xlim
            if xlim is not None:
                data = data.loc[(data['time']>=xlim[0]) & (data['time']<=xlim[1]), :]

            mod_f = data.columns.tolist()
            mod_f.remove('time')
            data[modules[i]] = data[mod_f].mean(axis=1)

            _ComplexCurvePlot(data, prpDf, tmp_ax1, tmp_ax2, xlim=xlim, ylim=ylim,
                          ticks_fontsize=ticks_fontsize,
                          x_label=x_label if index_row==max_row[index_col] else '', 
                              y_label=y_label if index_col==0 else '', label_fontsize=label_fontsize,
                          curve_colors=['#e9ecec']*(data.shape[1]-1), dot_size=0, linewidth=1,
                          ident_colors=ident_colors)

            modsumDf = dataDict[modules[i]][['time','value']].copy() #data[['time',modules[i]]].copy()
            
            _ComplexCurvePlot(modsumDf, prpDf, tmp_ax1, None, xlim=xlim, ylim=ylim,
                          ticks_fontsize=ticks_fontsize,
                          x_label=x_label if index_row==max_row[index_col] else '', 
                              y_label=y_label if index_col==0 else '', label_fontsize=label_fontsize,
                          curve_colors=[curve_colors[i]], dot_size=0, linewidth=linewidth,
                          ident_colors=ident_colors)


    else:

        # Setup the plot
        fig1 = plt.figure(constrained_layout=False)
        gs1 = fig1.add_gridspec(nrows=10, ncols=10, left=0.0, right=1.0, top=0.9, bottom=0.3, wspace=0.05, hspace=0.2)
        f1_ax1 = fig1.add_subplot(gs1[:-1, :-1])
        f1_ax2 = fig1.add_subplot(gs1[-1, :-1])
        f1_legend = fig1.add_subplot(gs1[:-1, -1])
        hfont = {'fontname':'sans'}
        fig1.suptitle(title, fontsize=title_fontsize, x=0.45 , y=0.95, verticalalignment='bottom', **hfont)

        _ComplexCurvePlot(dataDict, prpDf, f1_ax1, f1_ax2, xlim=xlim, ylim=ylim,
                          ticks_fontsize=ticks_fontsize,
                          x_label=x_label, y_label=y_label, label_fontsize=label_fontsize,
                          curve_colors=curve_colors, dot_size=0, linewidth=linewidth,
                          ident_colors=ident_colors)

    # Legend
    if show_legend:
        legend_elements = [Line2D([0], [0], linewidth=0, label='Curves')]
        legend_elements += [Line2D([0], [0], linewidth=0, marker='o', color=curve_colors[i], label=modules[i], markersize=7) for i in range(len(modules))]
        legend_elements += [Line2D([0], [0], linewidth=0, label='')]
        legend_elements += [Line2D([0], [0], linewidth=0, label=ident)]
        legend_elements += [Line2D([0], [0], linewidth=0, marker='o', color=ident_colors[i], label=path[i], markersize=7) for i in range(len(path))]
        #ax.set_title(feature)
        f1_legend.legend(handles=legend_elements, loc='center left', bbox_to_anchor=(0.1, 0.5), frameon=False)
        f1_legend.tick_params(left = False, right = False , labelleft = False ,
                           labelbottom = False, bottom = False)
        f1_legend.grid(which='both', linestyle='-', linewidth='0')
        f1_legend.axis('off')

    if save is None:
        plt.show()
    else:
        with plt.rc_context():
            plt.savefig(save)









import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="treasmo",
    version="1.0.5",
    author="Chaozhong Liu",
    author_email="czliubioinfo@gmail.com",
    description="Transcription regulation analysis toolkit for single-cell multi-omics data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ChaozhongLiu/TREASMO",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ),
)
